@@ -107,7 +107,7 @@ destroy_dialog (GtkWidget *w, gpointer data)
 }
 
 GtkHTMLEditPropertiesDialog *
-gtk_html_edit_properties_dialog_new (GtkHTMLControlData *cd, gboolean insert, gchar *title)
+gtk_html_edit_properties_dialog_new (GtkHTMLControlData *cd, gboolean insert, gchar *title, gchar *icon_path)
 {
 	GtkHTMLEditPropertiesDialog *d = g_new (GtkHTMLEditPropertiesDialog, 1);
 	GtkWindow *parent;
@@ -139,6 +139,8 @@ gtk_html_edit_properties_dialog_new (GtkHTMLControlData *cd, gboolean insert, gc
 	gnome_dialog_button_connect (GNOME_DIALOG (d->dialog), insert ? 1 : 2, prop_close, d);
 	gnome_dialog_set_default (GNOME_DIALOG (d->dialog), 0);
 
+	
+	gnome_window_icon_set_from_file (GTK_WINDOW (d->dialog), icon_path);
 	gnome_dialog_set_sensitive (GNOME_DIALOG (d->dialog), 0, FALSE);
 	if (!insert)
 		gnome_dialog_set_sensitive (GNOME_DIALOG (d->dialog), 1, FALSE);
