@@ -22,6 +22,9 @@
 #ifndef _GTKHTML_PRIVATE_H
 #define _GTKHTML_PRIVATE_H
 
+#include <libgnome/gnome-paper.h>
+#include <libgnomeprint/gnome-print.h>
+#include <libgnomeprint/gnome-print-master.h>
 #include <gtk/gtkwidget.h>
 #include "gtkhtml-types.h"
 #include "htmltypes.h"
@@ -38,7 +41,7 @@ struct _GtkHTMLPrivate {
 
 	gboolean update_styles;
 
-	gint last_selection_type;
+	gint selection_type;
 	/* Used to hold the primary selection when
 	** pasting within ourselves
 	*/
@@ -46,28 +49,40 @@ struct _GtkHTMLPrivate {
 	guint       primary_len;
 
 	gchar *content_type;
+	char  *base_url;
 
 	GtkWidget *search_input_line;
 
+	GnomePrintMaster *print_master;
+
 #ifdef GTKHTML_HAVE_GCONF
 	guint set_font_id;
 	guint notify_id;
+	guint notify_spell_id;
 #endif
 #ifdef GTK_HTML_USE_XIM
 	GdkICAttr *ic_attr;
 	GdkIC *ic;
 #endif
+
+	HTMLObject *dnd_object;
+	HTMLObject *dnd_real_object;
+	gboolean    dnd_in_progress;
+	gchar      *dnd_url;
+
+	guint32     event_time;
+	gboolean    selection_as_cite;
 };
 
 void  gtk_html_private_calc_scrollbars  (GtkHTML                 *html,
 					 gboolean                *changed_x,
 					 gboolean                *changed_y);
 void  gtk_html_editor_event_command     (GtkHTML                 *html,
-					 GtkHTMLCommandType       com_type);
+					 GtkHTMLCommandType       com_type,
+					 gboolean                 before);
 void  gtk_html_editor_event             (GtkHTML                 *html,
 					 GtkHTMLEditorEventType   event,
 					 GtkArg                 **args);
+void  gtk_html_api_set_language         (GtkHTML                 *html);
 
 #endif /* _GTKHTML_PRIVATE_H */
-
-
