@@ -146,20 +146,20 @@ copy (HTMLObject *self, HTMLObject *dest)
 	copy_sized (self, dest, HTML_TABLE (self)->totalRows, HTML_TABLE (self)->totalCols);
 }
 
-static HTMLObject * op_copy (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, guint *len);
+static HTMLObject * op_copy (HTMLObject *self, HTMLObject *parent, HTMLEngine *e, GList *from, GList *to, guint *len);
 
 static HTMLObject *
-copy_as_leaf (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, guint *len)
+copy_as_leaf (HTMLObject *self, HTMLObject *parent, HTMLEngine *e, GList *from, GList *to, guint *len)
 {
 	if ((!from || GPOINTER_TO_INT (from->data) == 0)
 	    && (!to || GPOINTER_TO_INT (to->data) == html_object_get_length (self)))
-		return op_copy (self, e, NULL, NULL, len);
+		return op_copy (self, parent, e, NULL, NULL, len);
 	else
 		return html_engine_new_text_empty (e);
 }
 
 static HTMLObject *
-op_copy (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, guint *len)
+op_copy (HTMLObject *self, HTMLObject *parent, HTMLEngine *e, GList *from, GList *to, guint *len)
 {
 	HTMLTableCell *start, *end;
 	HTMLTable *nt, *t;
@@ -170,7 +170,7 @@ op_copy (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, guint *len)
 	if ((from || to)
 	    && (!from || !from->next)
 	    && (!to || !to->next))
-		return copy_as_leaf (self, e, from, to, len);
+		return copy_as_leaf (self, parent, e, from, to, len);
 
 	t  = HTML_TABLE (self);
 	nt = g_new0 (HTMLTable, 1);
@@ -199,7 +199,7 @@ op_copy (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, guint *len)
 				if (cell->row == r + start->row && cell->col == c + start_col) {
 					HTMLTableCell *cell_copy;
 					cell_copy = HTML_TABLE_CELL
-						(html_object_op_copy (HTML_OBJECT (cell), e,
+						(html_object_op_copy (HTML_OBJECT (cell), HTML_OBJECT (nt), e,
 								      html_object_get_bound_list (HTML_OBJECT (cell), from),
 								      html_object_get_bound_list (HTML_OBJECT (cell), to),
 								      len));
@@ -389,6 +389,13 @@ split (HTMLObject *self, HTMLEngine *e, HTMLObject *child, gint offset, gint lev
 	dup_cell  = HTML_TABLE_CELL ((*right)->data);
 	cell      = HTML_TABLE_CELL ((*left)->data);
 
+	if (dup_cell->row == t->totalRows - 1 && dup_cell->col == t->totalCols - 1 && cell_is_empty (dup_cell)) {
+		dup = html_engine_new_text_empty (e);
+		html_object_destroy ((*right)->data);
+		g_list_free (*right);
+		*right = NULL;
+	} else {
+
 #ifdef GTKHTML_DEBUG_TABLE
 	printf ("before split\n");
 	printf ("-- self --\n");
@@ -449,7 +456,7 @@ split (HTMLObject *self, HTMLEngine *e, HTMLObject *child, gint offset, gint lev
 			}
 		}
 	}
-
+	}
 	html_clue_append_after (HTML_CLUE (self->parent), dup, self);
 
 	*left  = g_list_prepend (*left, self);
@@ -978,7 +985,7 @@ html_table_set_cells_position (HTMLTable *table, HTMLPainter *painter)
 				HTML_OBJECT (cell)->x = COLUMN_OPT (table, c) + pixel_size * border_extra;
 				HTML_OBJECT (cell)->y = ROW_HEIGHT (table, rl) + pixel_size * (- table->spacing)
 					- HTML_OBJECT (cell)->descent;
-				html_object_set_max_ascent (HTML_OBJECT (cell), painter,
+				html_object_set_max_height (HTML_OBJECT (cell), painter,
 							    ROW_HEIGHT (table, rl) - ROW_HEIGHT (table, cell->row)
 							    - pixel_size * (table->spacing + border_extra));
 			}
@@ -1006,6 +1013,13 @@ add_clear_area (GList **changed_objs, HTMLObject *o, gint x, gint w)
 	*changed_objs = g_list_prepend (*changed_objs, NULL);
 }
 
+static void
+html_table_set_max_height (HTMLObject *o, HTMLPainter *painter, gint height)
+{
+	/* for now just remember it, it will be passed down once size is calculated */
+	HTML_TABLE (o)->max_height = height;
+}
+
 static gboolean
 calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs)
 {
@@ -1126,7 +1140,7 @@ draw (HTMLObject *o,
 	ty += o->y - o->ascent;
 
 	/* Draw the cells */
-	get_bounds (table, x, y, width, height, &start_col, &end_col, &start_row, &end_row);
+	get_bounds (table, x - o->x, y - o->y + o->ascent, width, height, &start_col, &end_col, &start_row, &end_row);
 	for (r = start_row; r <= end_row; r++) {
 		for (c = start_col; c <= end_col; c++) {
 			cell = table->cells[r][c];
@@ -1422,9 +1436,9 @@ divide_upto_preferred_width (HTMLTable *table, HTMLPainter *painter, GArray *pre
 					- pixel_size * (table->spacing + border_extra);
 				if (max_size [c] < pw) {
 					processed_pw += pw;
-					part = (LL min_fill * processed_pw) / min_pw;
-					if (LL min_fill * processed_pw - LL part * min_pw
-					    > LL (part + 1) * min_pw - LL min_fill * processed_pw)
+					part = (LL min_fill * processed_pw) / total_fill;
+					if (LL min_fill * processed_pw - LL part * total_fill
+					    > LL (part + 1) * total_fill - LL min_fill * processed_pw)
 						part ++;
 					part         -= added;
 					added        += part;
@@ -1867,6 +1881,7 @@ fit_line (HTMLObject *o,
 	  HTMLPainter *painter,
 	  gboolean start_of_line,
 	  gboolean first_run,
+	  gboolean next_to_floating,
 	  gint width_left)
 {
 	return HTML_FIT_COMPLETE;
@@ -2159,6 +2174,7 @@ html_table_class_init (HTMLTableClass *klass,
 	object_class->calc_min_width = calc_min_width;
 	object_class->calc_preferred_width = calc_preferred_width;
 	object_class->set_max_width = html_table_set_max_width;
+	object_class->set_max_height = html_table_set_max_height;
 	object_class->reset = reset;
 	object_class->check_point = check_point;
 	object_class->find_anchor = find_anchor;
@@ -2274,17 +2290,21 @@ html_table_end_row (HTMLTable *table)
 	table->row++;
 }
 
-void
+gint
 html_table_end_table (HTMLTable *table)
 {
-	gint r, c;
+	gint r, c, cells = 0;
 
 	for (r = 0; r < table->totalRows; r ++)
 		for (c = 0; c < table->totalCols; c ++)
-			if (table->cells [r][c] && HTML_CLUE (table->cells [r][c])->head == NULL) {
-				HTMLTableCell *cell = table->cells [r][c];
+			if (table->cells [r][c]) {
+				if (HTML_CLUE (table->cells [r][c])->head == NULL) {
+					HTMLTableCell *cell = table->cells [r][c];
 
-				remove_cell (table, cell);
-				html_object_destroy (HTML_OBJECT (cell));
+					remove_cell (table, cell);
+					html_object_destroy (HTML_OBJECT (cell));
+				} else
+					cells ++;
 			}
+	return cells;
 }
